/*                                                           08-17-94 10:28pm
 ****************************************************************************
 *
 *   BGF.C            BGI CHR-fonts handling package (BGF).
 *                    This version of BGF package uses Zortech FG.
 *   Version 1.00.01
 *
 * Copyright (c) Alexander V. Loukyanets, 1994.  All rights reserved.
 *
 ****************************************************************************
 */

#define _BGF_C_Module  "Version 1.00.01"
#define _BGF_C_VrDate  "17.08.94"


#include <stdlib.h>
#include <stdio.h>
#include <dos.h>
#include <fg.h>
#include "bgf.h"


#define BGF_MAX_FILNAME_L   128
#define BGF_MAX_FNTNAME_L   4
#define BGF_MAX_FNTVERS_L   3
#define BGF_MAXSYMBINFONT   256
#define BGF_MAX_FFILEID_L   2
#define BGF_MAX_COPRGHT_L   253
#define BGF_COPRGHT_END_M   '\x1A'
#define BGF_FONT_HEADER_M   '+'


/*
 ****************************************************************************
 *   Types definitions
 */

typedef struct {
  int             FontIndex;
  unsigned char   FileName    [BGF_MAX_FILNAME_L+1];
  unsigned char   FontName    [BGF_MAX_FNTNAME_L+1];
  signed char     FontVers    [BGF_MAX_FNTVERS_L+1];
  short           CharsNum;
  unsigned char   CharStart;
  signed char     FillFlag;
  signed char     UprMargin;
  signed char     LwrMargin;
  short          *Pointers;
  signed char    *Widths;
  signed char    *FontTable;
} t_BGFFontStruct;


/*
 ****************************************************************************
 *   Internal Data
 */

char *BGF_Copyright="BGF Package "_BGF_C_Module" ("_BGF_C_VrDate") "
        " Copyright (c) Alexander V. Loukyanets, 1994.  All rights reserved.";

static t_BGFFontStruct      *RegdFonts;
static int                  NRegdFonts = 0;
static int                  Src_X    = 0,
                            Src_Y    = 0,
                            Src_Font,
                            Src_Color;
static double               Src_YCoef,
                            Src_XCoef;

/*
 ****************************************************************************
 *   Internal Functions Prototypes
 */

static void     DrawChar    ( unsigned char ch );
static int      PairXY      ( char x, char y );

//int             bgfUnregisterFont   ( int fontIndex );


/***************************************************************************/
/***************************************************************************/
/***************************************************************************/

/*
 ****************************************************************************
 */

int bgfRegisterFont( char *FileName )
{
  FILE           *BGFf;
  int             i;
  unsigned char   ch;
  char            buffer[256];
  short           TableSize;

  short           HeaderOffs;
  unsigned char   FontName    [BGF_MAX_FNTNAME_L+1];
  short           FontSize;
  signed char     FontVers    [BGF_MAX_FNTVERS_L+1];
  short           CharsNum;
  unsigned char   CharStart;
  short           DefOffs;
  signed char     FillFlag;
  signed char     UprMargin;
  signed char     LwrMargin;
  short          *Pointers=NULL;
  signed char    *Widths=NULL;
  signed char    *FontTable=NULL;

#define FREEMEM   { if ( Pointers != NULL )   free( Pointers );\
                    if ( Widths != NULL )     free( Widths );\
                    if ( FontTable != NULL )  free( FontTable );}

  if ( (BGFf=fopen( FileName, "rb" ))==NULL ) return( BGFE_OPNERRFNTFIL );
#define RETURN(i) { fclose(BGFf); FREEMEM; return(i); }
#define RETurn(i) {               FREEMEM; return(i); }

  //***FontFileID
  memset(buffer,0,BGF_MAX_FFILEID_L+1);
  if ( (fread(buffer,1,BGF_MAX_FFILEID_L,BGFf)) != BGF_MAX_FFILEID_L )
       RETURN( BGFE_REAERRFNTFIL );
  if ( strncmp(buffer,"PK",BGF_MAX_FFILEID_L) != 0 )
       RETURN( BGFE_ITSNOTFNTFIL );

  //***Copyright
  memset(buffer,0,BGF_MAX_COPRGHT_L+1);  i=0;
  while ( ( fread(&ch,1,1,BGFf) == 1 ) &&
          ( ch != BGF_COPRGHT_END_M ) &&
          ( i <=  BGF_MAX_COPRGHT_L ) ) buffer[i++]=ch;
  if ( ferror(BGFf) || i>BGF_MAX_COPRGHT_L )
       RETURN( BGFE_REAERRFNTFIL );

  //***HeaderOffs
  if ( (fread(&HeaderOffs,2,1,BGFf)) != 1 )
       RETURN( BGFE_REAERRFNTFIL );

  //***FontName
  memset(FontName,0,BGF_MAX_FNTNAME_L+1);
  if ( (fread(FontName,1,BGF_MAX_FNTNAME_L,BGFf)) != BGF_MAX_FNTNAME_L )
       RETURN( BGFE_REAERRFNTFIL );

  //***FontSize
  if ( (fread(&FontSize,2,1,BGFf))!=1 )
       RETURN( BGFE_REAERRFNTFIL );

  //***FontVers
  memset(FontVers,0,BGF_MAX_FNTVERS_L+1);
  if ( (fread(FontVers,1,BGF_MAX_FNTVERS_L,BGFf)) != BGF_MAX_FNTVERS_L )
       RETURN( BGFE_REAERRFNTFIL );

  //***Seek File on HeaderOffset
  if ( (fseek(BGFf,(long)HeaderOffs,SEEK_SET)) != 0 )
       RETURN( BGFE_REAERRFNTFIL );

  //***FontHeadMarker
  if ( (fread(&ch,1,1,BGFf)) != 1 )
       RETURN( BGFE_REAERRFNTFIL );
  if ( ch != BGF_FONT_HEADER_M )
       RETURN( BGFE_ITSNOTFNTFIL );

  //***CharsNum
  if ( (fread(&CharsNum,2,1,BGFf)) != 1 )
       RETURN( BGFE_REAERRFNTFIL );
  if ( CharsNum >= BGF_MAXSYMBINFONT )
       RETURN( BGFE_ITSNOTFNTFIL );

  //***Skip Dummy1
  if ( (fread(&ch,1,1,BGFf)) != 1 )
       RETURN( BGFE_REAERRFNTFIL );

  //***CharsStart
  if ( (fread(&CharStart,1,1,BGFf)) != 1 )
       RETURN( BGFE_REAERRFNTFIL );

  //***DefOffs
  if ( (fread(&DefOffs,2,1,BGFf)) != 1 )
       RETURN( BGFE_REAERRFNTFIL );

  //***FillFlag
  if ( (fread(&FillFlag,1,1,BGFf)) != 1 )
       RETURN( BGFE_REAERRFNTFIL );

  //***UprMargin
  if ( (fread(&UprMargin,1,1,BGFf)) != 1 )
       RETURN( BGFE_REAERRFNTFIL );

  //***Skip Dummy2
  if ( (fread(&ch,1,1,BGFf)) != 1 )
       RETURN( BGFE_REAERRFNTFIL );

  //***LwrMargin
  if ( (fread(&LwrMargin,1,1,BGFf)) != 1 )
       RETURN( BGFE_REAERRFNTFIL );

  //***Skip Dummy3
  if ( (fread(buffer,1,5,BGFf)) != 5 )
       RETURN( BGFE_REAERRFNTFIL );

  //***Pointers
  if ( (Pointers=(short*)malloc( 2*CharsNum*sizeof(short) )) == NULL )
       RETURN( BGFE_MEMALLOC_ERR );
  if ( (fread(Pointers,2,CharsNum,BGFf)) != CharsNum )
       RETURN( BGFE_REAERRFNTFIL );

  //***Widths
  if ( (Widths  = (signed char*)malloc(   CharsNum*sizeof(signed char)  )) == NULL )
       RETURN( BGFE_MEMALLOC_ERR );
  if ( (fread(Widths,1,CharsNum,BGFf)) != CharsNum )
       RETURN( BGFE_REAERRFNTFIL );

  //***Seek File on FontTable
  if ( (fseek(BGFf,(long)(HeaderOffs+DefOffs),SEEK_SET))!=0 )
       RETURN( BGFE_REAERRFNTFIL );

  //TableSize = FontSize-16-3*CharsNum;
  TableSize = FontSize-DefOffs;

  //***FontTable
  if ( (FontTable=(signed char*)malloc( TableSize*sizeof(signed char) )) == NULL )
       RETURN( BGFE_MEMALLOC_ERR );
  if ( (fread(FontTable,1,TableSize,BGFf)) != TableSize )
       RETURN( BGFE_REAERRFNTFIL );

  if( ferror(BGFf) )
       RETURN( BGFE_REAERRFNTFIL );

  fclose( BGFf );
#undef RETURN

  if ( NRegdFonts == 0 ) {
    if ( NULL == (RegdFonts = (t_BGFFontStruct*)malloc(sizeof(t_BGFFontStruct))) )
       RETurn( BGFE_MEMALLOC_ERR );
  }
  else {
    if ( NULL == (RegdFonts = (t_BGFFontStruct*)realloc(RegdFonts,(NRegdFonts+1)*sizeof(t_BGFFontStruct))) )
       RETurn( BGFE_MEMALLOC_ERR );
  }
#undef RETurn
#undef FREEMEM

  RegdFonts[NRegdFonts].FontIndex = NRegdFonts+1;
  strncpy(
  RegdFonts[NRegdFonts].FileName,     FileName, BGF_MAX_FILNAME_L+1);
  strncpy(
  RegdFonts[NRegdFonts].FontName,     FontName, BGF_MAX_FNTNAME_L+1);
  for ( i=0; i<BGF_MAX_FNTVERS_L; i++)
  RegdFonts[NRegdFonts].FontVers[i] = FontVers[i];
  RegdFonts[NRegdFonts].CharsNum    = CharsNum;
  RegdFonts[NRegdFonts].CharStart   = CharStart;
  RegdFonts[NRegdFonts].FillFlag    = FillFlag;
  RegdFonts[NRegdFonts].UprMargin   = UprMargin;
  RegdFonts[NRegdFonts].LwrMargin   = LwrMargin;
  RegdFonts[NRegdFonts].Pointers    = Pointers;
  RegdFonts[NRegdFonts].Widths      = Widths;
  RegdFonts[NRegdFonts].FontTable   = FontTable;

  return( ++NRegdFonts );
}


/*
 ****************************************************************************
 */

int bgfRegisterFonts( char *FileNames )
{
  struct FIND     ffblk;
  int             status, i=0;

  if ( 0 == _dos_findfirst( FileNames, 0xff , &ffblk ) ) {
    do {
      i++;
      status=bgfRegisterFont( ffblk.name );
    } while ( status >= 0  &&  0 == _dos_findnext( &ffblk ) );
  }
  return( (status>=0) ? i : status );
}


/*
 ****************************************************************************
 */

/***
int bgfUnregisterFont( int fontIndex )
{

  return( BGFE_OK );
}
***/


/*
 ****************************************************************************
 */

int bgfUnregisterFonts( void )
{
  int i=0;

  if ( NRegdFonts <= 0 ) return( BGFE_NO_REGDFONTS );
  for ( ; i < NRegdFonts; i++) {
    free( RegdFonts[i].Pointers );
    free( RegdFonts[i].Widths );
    free( RegdFonts[i].FontTable );
  }
  free( RegdFonts );
  NRegdFonts = 0;
  return( BGFE_OK );
}


/*
 ****************************************************************************
 */

int bgfDrawText( int fontIndex,
                 int color,
                 int x, int y,
                 int Height, double ExtCoef,
                 char *Text )
{
  register int i=0;

  if ( NRegdFonts == 0 ) return( BGFE_NO_REGDFONTS );
  while ( fontIndex != RegdFonts[i].FontIndex  &&  i < NRegdFonts ) i++;
  if ( i == NRegdFonts ) return( BGFE_NO_SUCH_FONT );

  Src_Font  = i;
  Src_X     = x;
  Src_Y     = y;
  Src_Color = color;
  Src_YCoef = ((double)Height)/((double)(RegdFonts[i].UprMargin
                                        -RegdFonts[i].LwrMargin));
  Src_XCoef = (ExtCoef==0.0) ? Src_YCoef : ExtCoef*Src_YCoef;

  while ( (*Text) != '\0' ) {
    DrawChar( *Text );
    Text++;
  }

  return( BGFE_OK );
}


/*
 ****************************************************************************
 */

static void DrawChar( unsigned char ch )
{
  register int i;
  char         x, y;

  if ( (i=ch-RegdFonts[Src_Font].CharStart) < RegdFonts[Src_Font].CharsNum ) {
    i=RegdFonts[Src_Font].Pointers[i];
    do {
      x=RegdFonts[Src_Font].FontTable[i++];
      y=RegdFonts[Src_Font].FontTable[i++];
    } while ( PairXY( x, y ) );
  }
}


/*
 ****************************************************************************
 */

static int PairXY( char x, char y )
{

#define    Is_HIBit(byte)  (byte & 0x80)
#define Clear_HIBit(byte)  (byte & 0x7F)
#define  Correction(byte)  if ( (byte) >= 63 ) byte-=128

  char              Hx=Is_HIBit(x),
                    Hy=Is_HIBit(y);
  register int      ret=0;
  static fg_line_t  line;

  if ( Hx ) {   x=Clear_HIBit(x);
    if ( Hy ) { y=Clear_HIBit(y); ret=3; }
    else                          ret=2;
  }
  else
    if ( Hy ) { y=Clear_HIBit(y); ret=1; }

  Correction(x);
  Correction(y);

  switch ( ret )
  {
    case 0 : //*** End of Symbol definition
              Src_X = line[FG_X1];
              Src_Y = line[FG_Y1];
              break;
    case 2 : //*** Start New Segment
              line[FG_X1] = Src_X + x*Src_XCoef;
              line[FG_Y1] = Src_Y + y*Src_YCoef;
              break;
    case 1 : //***
              break;
    case 3 : //*** Normal Point
              line[FG_X2] = Src_X + x*Src_XCoef;
              line[FG_Y2] = Src_Y + y*Src_YCoef;
              fg_drawlineclip( Src_Color, FG_MODE_SET, ~0, FG_LINE_SOLID,
                               line,      fg.displaybox );
              line[FG_X1] = line[FG_X2];
              line[FG_Y1] = line[FG_Y2];
              break;
  }

  return( ret );

#undef    Is_HIBit
#undef Clear_HIBit
#undef Correction
}


/*
 ****************************************************************************
 */

int bgfGetNRegdFonts( void )
{
  return( NRegdFonts );
}


/*
 ****************************************************************************
 */

int bgfGetFontInfo( int fontIndex, t_BGFFontInfo *fi )
{
  register int i=0;

  if ( NRegdFonts == 0 ) return( BGFE_NO_REGDFONTS );
  while ( fontIndex != RegdFonts[i].FontIndex  &&  i < NRegdFonts ) i++;
  if ( i == NRegdFonts ) return( BGFE_NO_SUCH_FONT );

  fi->FileName  = RegdFonts[i].FileName;
  fi->FontName  = RegdFonts[i].FontName;
  fi->FontVers  = RegdFonts[i].FontVers;
  fi->CharsNum  = RegdFonts[i].CharsNum;
  fi->CharStart = RegdFonts[i].CharStart;
  fi->FillFlag  = RegdFonts[i].FillFlag;
  fi->UprMargin = RegdFonts[i].UprMargin;
  fi->LwrMargin = RegdFonts[i].LwrMargin;

  return( BGFE_OK );
}


/****************************************************************************
*** end of BGI_CHR.C ***/